package server;

import client.Client;
import enums.FileType;
import enums.MethodEnum;
import enums.ReturnCodes;
import org.blinkenlights.jid3.ID3Exception;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import servlet.ProcessAction;
import util.Constants;
import util.Util;

import java.io.*;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URLDecoder;
import java.util.StringTokenizer;

public class Server extends Thread {

    private Client client;
    private int port;
    private static Log LOG = LogFactory.getLog(Server.class);

    public Server(int listen_port, Client client) {
        this.client = client;
        port = listen_port;
        this.start();
    }

    private void sendMessageToClient(String message) {
        LOG.info(message);
        client.sendMessage(message);
    }

    public void run() {
        LOG.info("running ");
        sendMessageToClient("Welcome to " + Constants.APP_NAME + "\n");
        ServerSocket serversocket;
        try {
            sendMessageToClient("Attempting to lauch on port " + Integer.toString(port) + "...");
            serversocket = new ServerSocket(port);
        }
        catch (Exception e) {
            handleError(null, "unable to launch: " + e.getMessage());
            return;
        }

        sendMessageToClient("OK!\n");
        sendMessageToClient("[ready, lets grab some MP3's!!!]\n\n");

        while (true) {
            try {
                Socket connectionsocket = serversocket.accept();
                InetAddress client = connectionsocket.getInetAddress();
                sendMessageToClient(client.getHostName() + " connected to server.\n");
                BufferedReader input = new BufferedReader(new InputStreamReader(connectionsocket.getInputStream()));
                String requestHeader = input.readLine();
                if (!"localhost".equalsIgnoreCase(client.getHostName()) && requestHeader.indexOf("../") != -1 && requestHeader.indexOf("..\\") != -1 ) {
                    throw new RuntimeException("Illegal access from client: " + client.getHostName() + "|" + client.getHostAddress() + " | request: " + requestHeader);
                }
                DataOutputStream output = new DataOutputStream(connectionsocket.getOutputStream());
                try {
                    try {
                        handleHTTPRequest(client.getHostAddress(), requestHeader, output);
                    } catch (Exception e) {
                        handleError(output, e.getMessage());
                        e.printStackTrace();
                    }
                } finally {
                    output.flush();
                    output.close();
                }
            }
            catch (Exception e) {
                handleError(null, e.getMessage());
                e.printStackTrace();
            }
        }
    }

    private void handleHTTPRequest(String requestId, String url, DataOutputStream output) throws ID3Exception, IOException {
        LOG.info("handling request");
        LOG.info("processing url: " + url);
        MethodEnum method = getRequestMethod(url, output);
        Request request = getRequestParameters(requestId, url);
        String action = request.getParamter(Constants.PARAM_ACTION);
        LOG.debug("processing request with action: " + action);
        if (action != null) {
            if (Constants.ACTION_GRAB_MP3.equals(action)) {
                try {
                    ProcessAction.handleMP3Request(output, method, request);
                } catch (Exception e) {
                    throw new RuntimeException("unable to rip MP3 " + e.getMessage());
                }
            }
            else if (Constants.ACTION_GET_BIO.equals(action)) {
                try {
                    ProcessAction.getArtistBio(request, output);
                } catch (Exception e) {
                     throw new RuntimeException("unable to retrieve bio from last.fm: " + e.getMessage());
                }
            }
            else if (Constants.ACTION_GET_STATS.equals(action)) {
                try {
                    ProcessAction.getStats(request, output);
                } catch (IOException e) {
                     throw new RuntimeException("unable to retrieve song stats from last.fm: " + e.getMessage());
                }
            }
            else if (Constants.ACTION_TOP_FRIENDS.equals(action)) {
                try {
                    ProcessAction.getTopFans(request, output);
                } catch (Exception e) {
                     throw new RuntimeException("unable to retrieve top fans from last.fm: " + e.getMessage());
                }
            }
            else if (Constants.ACTION_ADD_TRACK.equals(action)) {
                try {
                    ProcessAction.addTrackToLastFm(request, output);
                } catch (Exception e) {
                     throw new RuntimeException("unable add track to last.fm: " + e.getMessage());
                }
            }
            else {
                throw new RuntimeException("unsupported action: " + action);
            }
        } else {
            ProcessAction.handleStaticRequest(output, method, url);
        }

    }

    private Request getRequestParameters(String requestId, String requestString) throws UnsupportedEncodingException {
        Request request = new Request(requestId);
        requestString = URLDecoder.decode(requestString, "UTF-8");
        int beginIndex = requestString.indexOf("?");
        if (beginIndex == -1) {
            return request;
        }
        requestString = requestString.substring(beginIndex + 1, requestString.indexOf(" HTTP"));
        StringTokenizer params = new StringTokenizer(requestString, "&");

        LOG.debug("params found: " + params.countTokens());

        while (params.hasMoreTokens()) {
            String p = params.nextToken();
            StringTokenizer nv = new StringTokenizer(p, "=");
            if (nv.countTokens() != 2) {
                continue;
            }
            String name = nv.nextToken().trim();
            String value = nv.nextToken().trim();

            if (!name.equals("") && !value.equals("")) {
                request.addParameter(name, value);
                LOG.debug("adding param: [name: " + name + " value: " + value + "]");
            }

        }

        return request;
    }

    private MethodEnum getRequestMethod(String url, DataOutputStream output) {
        MethodEnum method = null;
        if (url.toUpperCase().startsWith("GET")) {
            method = MethodEnum.GET;
        } else if (url.toUpperCase().startsWith("HEAD")) {
            method = MethodEnum.HEAD;
        } else if (url.toUpperCase().startsWith("POST")) {
            method = MethodEnum.POST;
        } else {
            try {
                handleError(output, "unsupported request\n");
                throw new RuntimeException("unsupported request\n");
            }
            catch (Exception e) {
                handleError(output, e.getMessage());
            }
        }
        return method;
    }

    private void handleError(DataOutputStream output, String message) {
        sendMessageToClient(message + "\n");
        LOG.equals("error: " + message);
        if (output != null) {
            try {
                output.writeBytes(Util.writeResponseHeader(ReturnCodes.INTERNAL_ERROR, FileType.HTML, null));
                output.writeBytes(message);
            } catch (IOException e) {
                //ignore
            }
        }
    }

}
